import json
from typing import Dict
from numpy import add
from transformers import AutoModelForCausalLM, AutoTokenizer
import torch
from realign.utils.draw import create_image_from_list
import random
import PIL


def prepare_data(data_path: str):
    data = [json.loads(line) for line in open(data_path, 'r').readlines()]
    data = [dp for dp in data if '/people/person/nationality' in dp['relation']]
    return data

def prepare_model(
        base_model_path: str, base_model_device: str, 
        chat_model_path: str, chat_model_device: str,
        stylish_model_path: str, stylish_model_device: str,
):
    base_model, base_tokenizer = AutoModelForCausalLM.from_pretrained(base_model_path, device_map=base_model_device), AutoTokenizer.from_pretrained(base_model_path)

    chat_model, chat_tokenizer = AutoModelForCausalLM.from_pretrained(chat_model_path, device_map=chat_model_device), AutoTokenizer.from_pretrained(base_model_path)

    stylish_model, stylish_tokenizer = AutoModelForCausalLM.from_pretrained(stylish_model_path, device_map=stylish_model_device), AutoTokenizer.from_pretrained(base_model_path)

    base_tokenizer.pad_token_id = base_tokenizer.eos_token_id
    chat_tokenizer.pad_token_id = chat_tokenizer.eos_token_id
    stylish_tokenizer.pad_token_id = stylish_tokenizer.eos_token_id

    return base_model, base_tokenizer, chat_model, chat_tokenizer, stylish_model, stylish_tokenizer


def eval_shift(data: Dict, base_model, base_tokenizer, chat_model, chat_tokenizer, stylish_model, stylish_tokenizer):
    text: str = data['text']
    person: str = data['h']['name']
    nationality: str = data['t']['name']

    groundtruth: str = (
        "```json\n"
        "{\n"
        f"  \"person\": \"{person}\",\n"
        f"  \"nationality\": \"{nationality}\"\n"
        "}\n"
        "```"
    )

    groundtruth_ids = [base_tokenizer.bos_token_id] + base_tokenizer.encode(groundtruth, add_special_tokens=False) + [base_tokenizer.eos_token_id]
    groundtruth_ids = torch.tensor(
        [groundtruth_ids], dtype=torch.int64
    )

    # 1. chat model generate
    prompt_template = (
        "Below is an instruction that describes a task, paired with an input that provides further context. "
        "Write a response that appropriately completes the request.\n\n"
        "### Instruction:\nGiven a piece of text, please find out the person-nationality relation in it. Tell me who is the person and which is the nationality. The answer should be in json format.\n\n### Input:\n{text}\n\n### Response:"
    )
    prompt = prompt_template.format(text=text)

    def chat_model_generate():
        input_ids = chat_tokenizer.encode(prompt, return_tensors='pt', add_special_tokens=False).to(chat_model.device)
        generate_ids = chat_model.generate(input_ids, max_new_tokens=1024)
        # response = chat_tokenizer.decode(generate_ids[0][len(input_ids[0]):], skip_special_tokens=True)
        # return response
        return [generate_ids[0][len(input_ids[0]):].tolist()]

    # O: str = chat_model_generate()
    generate_output_ids = torch.tensor(
        chat_model_generate(), dtype=torch.int64
    )

    def get_rank(model, tokenizer, output_ids):
        prompt_ids = tokenizer.encode(prompt, return_tensors='pt', add_special_tokens=False)
        # output_ids = tokenizer.encode(O, return_tensors='pt')
        input_ids = torch.cat([prompt_ids, output_ids], dim=-1)
        input_ids = input_ids.to(model.device)
        with torch.no_grad():
            logits = model(input_ids, labels=input_ids)['logits']
        probabilities = torch.nn.functional.softmax(logits, dim=-1)
        ranks = []
        for i, token_id in enumerate(input_ids[0][len(prompt_ids[0]):], start=len(prompt_ids[0])-1):
            word = tokenizer.decode(token_id)
            prob = probabilities[0, i, token_id].item()

            # find the rank of token_id, in reverse order
            rank = (probabilities[0, i] > prob).sum().item()
            ranks.append((word, rank))
        return ranks

    return {
        "prompt": [(prompt, -1)],
        "base_model_generate": [("base_model_generate", -1)] + get_rank(base_model, base_tokenizer, generate_output_ids),
        "stylish_model_generate": [("stylish_model_generate", -1)] + get_rank(stylish_model, stylish_tokenizer, generate_output_ids),
        "base_model_groundtruth": [("base_model_groundtruth", -1)] + get_rank(base_model, base_tokenizer, groundtruth_ids),
        "stylish_model_groundtruth": [("stylish_model_groundtruth", -1)] + get_rank(stylish_model, stylish_tokenizer, groundtruth_ids),
    }



def main():
    data_path = 'data/nyt10/nyt10_test.txt'
    base_model_path = '/home/tushilong/hf/models/Llama-2-7b-hf'
    base_model_device = 'cuda:0'
    chat_model_path = 'ckpts/full'
    chat_model_device = 'cuda:1'
    stylish_model_path = 'ckpts/stylish'
    stylish_model_device = 'cuda:6'

    data = random.sample(prepare_data(data_path), 20)
    base_model, base_tokenizer, chat_model, chat_tokenizer, stylish_model, stylish_tokenizer = prepare_model(
        base_model_path, base_model_device, chat_model_path, chat_model_device, stylish_model_path, stylish_model_device
    )
    for idx, d in enumerate(data):
        results = eval_shift(d, base_model, base_tokenizer, chat_model, chat_tokenizer, stylish_model, stylish_tokenizer)
        
        images = {key: create_image_from_list(val) for key, val in results.items()}
        
        combine_width = max([img.width for img in images.values()])
        combine_height = sum([img.height for img in images.values()])
        combine_image = PIL.Image.new('RGB', (combine_width, combine_height), (255, 255, 255))
        y = 0
        for key in ["prompt", "base_model_generate", "stylish_model_generate", "base_model_groundtruth", "stylish_model_groundtruth"]:
            img = images[key]
            combine_image.paste(img, (0, y))
            y += img.height
        combine_image.save(f'outputs/{idx}.png')


if __name__ == "__main__":
    main()